@@ -22,7 +22,9 @@ use Apache::Const -compile => qw(DECLINED OK
 use Time::HiRes ();
 use MIME::Base64 ();
 
-our $VERSION = '1.1';
+use Apache::ClickPath::_parse ();
+
+our $VERSION = '1.2a';
 our $rcounter=int rand 0x10000;
 
 my @directives=
@@ -118,22 +120,11 @@ sub ClickPathUAExceptionsFile {
   $I->{"ClickPathUAExceptionsFile"}=$arg;
 }
 
-sub _parse_UAExceptions {
-  my $conf=shift;
-  my $a=[];
-  foreach my $line (split /\r?\n/, $conf) {
-    if( $line=~/^\s*(\w+):?\s+(.+?)\s*$/ ) {
-      push @{$a}, [$1, qr/$2/];
-    }
-  }
-  return $a;
-}
-
 sub ClickPathUAExceptions {
   my($I, $parms, @args)=@_;
 
   $I->{"ClickPathUAExceptions"}
-    =_parse_UAExceptions( $parms->directive->as_string );
+    =Apache::ClickPath::_parse::UAExceptions( $parms->directive->as_string );
 }
 
 sub ClickPathUAExceptionsEND {
@@ -146,46 +137,12 @@ sub ClickPathFriendlySessionsFile {
   $I->{"ClickPathFriendlySessionsFile"}=$arg;
 }
 
-sub _parse_FriendlySessions {
-  my $conf=shift;
-  my $t={};
-  my $r={};
-
-  foreach my $l (split /\r?\n/, $conf) {
-
-    next unless( $l=~/^\s*(\S+)\s+	# $1: friendly REMOTE_HOST
-                      (			# $2: list of "uri( number )" or
-                       (?:		#     "param( name )" statements
-		        (?:uri|param)\s*
-		        \(
-		          \s*\w+\s*
-                        \)\s*
-                       )+
-                      )
-                      (?:\s*(\w+))?	# $3: opt. name, default=REMOTE_HOST
-                     /x );
-
-    my ($rem_host, $stmt_list, $name)=($1, $2, $3);
-    $name=$rem_host unless( defined $name );
-
-    my @stmts;
-    while( $stmt_list=~/(uri|param)\s*\(\s*(\w+)\s*\)/g ) {
-      push @stmts, [$1, $2];
-    }
-
-    $t->{$rem_host}=[[@stmts], $name];
-    $r->{$name}=$rem_host;
-  }
-
-  return $t, $r;
-}
-
 sub ClickPathFriendlySessions {
   my($I, $parms, @args)=@_;
 
   @{$I}{"ClickPathFriendlySessionsTable",
 	"ClickPathFriendlySessionsReverse"}
-    =_parse_FriendlySessions( $parms->directive->as_string );
+    =Apache::ClickPath::_parse::FriendlySessions( $parms->directive->as_string );
 }
 
 sub ClickPathFriendlySessionsEND {
@@ -218,14 +175,14 @@ sub _get_ua_exc {
 
     local $/;
     $cf->{"ClickPathUAExceptions_list"}
-      =_parse_UAExceptions( scalar( <$fh> ) );
+      =Apache::ClickPath::_parse::UAExceptions( scalar( <$fh> ) );
     close $fh;
     return $cf->{"ClickPathUAExceptions_list"};
   } elsif( @stat and
 	   $stat[9]<=$cf->{"ClickPathUAExceptionsFile_read_time"} ) {
     return $cf->{"ClickPathUAExceptions_list"};
   } else {
-    return $cf->{"ClickPathUAExceptions"};
+    return $cf->{"ClickPathUAExceptions"} || [];
   }
 }
 
@@ -248,7 +205,7 @@ sub _get_friendly_session {
     local $/;
     @{$cf}{"ClickPathFriendlySessionsFileTable",
 	   "ClickPathFriendlySessionsFileReverse"}
-      =_parse_FriendlySessions( scalar( <$fh> ) );
+      =Apache::ClickPath::_parse::FriendlySessions( scalar( <$fh> ) );
     close $fh;
     return @{$cf}{"ClickPathFriendlySessionsFileTable",
 		  "ClickPathFriendlySessionsFileReverse"};
@@ -276,7 +233,15 @@ sub handler {
 
   my $file=$r->uri;
 
-  my $pr=$r->prev;
+  # if an old session is used this will be reset
+  # if an old session timed out (goto NEWSESSION) this will be incremented
+  # giving in 2.
+  # if simply a new session is created this is 1
+  # then a pnote is set indicating this state. Thus the filter can
+  # examin it.
+  my $newsession=1;
+
+  my $pr=$r->main || $r->prev;
   my $ref=$r->headers_in->{Referer} || "";
 
   if( $pr ) {
@@ -293,8 +258,9 @@ sub handler {
       $r->subprocess_env( SESSION_AGE=>
 			  $pr->subprocess_env( 'SESSION_AGE' ) );
       $r->subprocess_env( SESSION=>$session );
-      $r->pnotes( __PACKAGE__.'::newsession'=>1 )
-	if( $r->pnotes( __PACKAGE__.'::newsession' ) );
+      $newsession=$pr->pnotes( __PACKAGE__.'::newsession' );
+      $r->pnotes( __PACKAGE__.'::newsession'=>$newsession )
+	if( $newsession );
       #print STDERR "$$: ReUsing session $session\n";
     }
   } elsif( $file=~s!^/+\Q$tag\E ( [^/]+ ) /!/!x ) {
@@ -337,11 +303,13 @@ sub handler {
       $r->subprocess_env->unset( 'CGI_SESSION' );
       $r->subprocess_env->unset( 'REMOTE_SESSION' );
       $r->subprocess_env->unset( 'REMOTE_SESSION_HOST' );
+      $newsession++;
       goto NEWSESSION;
     } else {
       $r->subprocess_env( SESSION_START=>$l[0] );
       $r->subprocess_env( SESSION_AGE=>$r->request_time-$l[0] );
     }
+    $newsession=0;
   } else {
     $ref=~s!^(\w+://[^/]+)/\Q$tag\E[^/]+!$1!;
     $r->headers_in->{Referer}=$ref;
@@ -428,7 +396,7 @@ sub handler {
       $r->subprocess_env( SESSION_START=>$r->request_time );
       $r->subprocess_env( SESSION_AGE=>0 );
       $r->subprocess_env( CGI_SESSION=>'/'.$tag.$session );
-      $r->pnotes( __PACKAGE__.'::newsession'=>1 );
+      $r->pnotes( __PACKAGE__.'::newsession'=>$newsession );
       #print STDERR "$$: Using new session $session\n";
     }
   }
@@ -462,12 +430,19 @@ sub OutputFilter {
 
     $host=$r->headers_in->{Host};
 
-    if( $r->pnotes( __PACKAGE__.'::newsession' ) ) {
+    my $newsession=$r->pnotes( __PACKAGE__.'::newsession' );
+    if( $newsession ) {
       $the_request=$r->the_request;
       $the_request=~s/^\s*\w+\s+//;
       $the_request=~s![^/]*[\s?].*$!!;
+      #print STDERR "the_request=$the_request\n";
+      if( $newsession==2 ) {
+	# cut off an timed out old session if given
+	$the_request=~s!^/+\Q$sprefix\E[^/]+!!;
+	#print STDERR "the_request(2)=$the_request\n";
+      }
 
-      my $re=qr,^(https?://\Q$host\E)?(?!\w+://)(.),i;
+      my $re=qr,^(https?://\Q$host\E)?(?!\w+:)(.),i;
       $r->headers_out->{Location}=~s!$re!$2 eq '/'
                                          ? $1.$sess.$2
                                          : $1.$sess.$the_request.$2
@@ -479,7 +454,7 @@ sub OutputFilter {
                                             !e
 	if( exists $r->err_headers_out->{Location} );
 
-      $re=qr,^(\s*\d+\s*;\s*url\s*=\s*(?:https?://\Q$host\E)?)(?!\w+://)(.),i;
+      $re=qr,^(\s*\d+\s*;\s*url\s*=\s*(?:https?://\Q$host\E)?)(?!\w+:)(.),i;
       $r->headers_out->{Refresh}=~s!$re!$2 eq '/'
                                         ? $1.$sess.$2
                                         : $1.$sess.$the_request.$2
@@ -525,7 +500,7 @@ sub OutputFilter {
 	      )			# Das alles ist in $1
 	      (?:/+\Q$sprefix\E[^/]+)?
 	      (			# $3 start
-	       (?!\w+://).*?	# ein beliebiger nicht mit http:// o.ae.
+	       (?!(?:\w+:|\043)).*? # ein beliebiger nicht mit http:// o.ae.
 				#   beginnender String (moeglichst kurz)
 	       \2		# das schließende Quote: $2
 	      )			# $3 ende
@@ -539,7 +514,7 @@ sub OutputFilter {
 	      )			# $1 ende
 
 	      (			# $2 start         optional "http-equiv=..."
-	       \bhttp-equiv\s*=\s*(["'])refresh\3
+	       \bhttp-equiv\s*=\s*(["']?)refresh\3
 	       [^>]*?		# evtl. anderes Zeug
 	      )?		# $2 ende
 
@@ -553,14 +528,14 @@ sub OutputFilter {
 	      (?:/+\Q$sprefix\E[^/]+)?
 
 	      (			# $6 start         URL + closing quote
-	       (?!\w+://).*?	# ein beliebiger nicht mit http:// o.ae.
+	       (?!\w+:).*?	# ein beliebiger nicht mit http:// o.ae.
 				#   beginnender String (so kurz wie möglich)
 	       \5		# das schließende Quote: $5
 	      )			# $6 ende
 
 	      (			# $7 start         optional "http-equiv=..."
 	       [^>]*?		# evtl. anderes Zeug
-	       \bhttp-equiv\s*=\s*(["'])refresh\8
+	       \bhttp-equiv\s*=\s*(["']?)refresh\8
 	      )?		# $7 ende
 	     ,ix;
 
@@ -573,7 +548,7 @@ sub OutputFilter {
 	      )			# Das alles ist in $1
 	      (?:/+\Q$sprefix\E[^/]+)?
 	      (			# $3 start
-	       (?!\w+://).*?	# ein beliebiger nicht mit http:// o.ae.
+	       (?!\w+:).*?	# ein beliebiger nicht mit http:// o.ae.
 				#   beginnender String (moeglichst kurz)
 	       \2		# das schließende Quote: $2
 	      )			# $3 ende
@@ -588,7 +563,7 @@ sub OutputFilter {
 	      )			# Das alles ist in $1
 	      (?:/+\Q$sprefix\E[^/]+)?
 	      (			# $3 start
-	       (?!\w+://).*?	# ein beliebiger nicht mit http:// o.ae.
+	       (?!\w+:).*?	# ein beliebiger nicht mit http:// o.ae.
 				#   beginnender String (moeglichst kurz)
 	       \2		# das schließende Quote: $2
 	      )			# $3 ende
@@ -617,7 +592,7 @@ sub OutputFilter {
 	      )			# $1 ende
 
 	      (			# $2 start         optional "http-equiv=..."
-	       \bhttp-equiv\s*=\s*(["'])refresh\3
+	       \bhttp-equiv\s*=\s*(["']?)refresh\3
 	       [^>]*?		# evtl. anderes Zeug
 	      )?		# $2 ende
 
@@ -638,7 +613,7 @@ sub OutputFilter {
 
 	      (			# $7 start         optional "http-equiv=..."
 	       [^>]*?		# evtl. anderes Zeug
-	       \bhttp-equiv\s*=\s*(["'])refresh\8
+	       \bhttp-equiv\s*=\s*(["']?)refresh\8
 	      )?		# $7 ende
 	     ,ix;
 
@@ -703,7 +678,6 @@ sub OutputFilter {
 
   # now, filter the content
   while( $f->read(my $buffer, 1024) ) {
-
     # prepend any tags leftover from the last buffer or invocation
     $buffer = $context->{extra} . $buffer if( length $context->{extra} );
 
@@ -1083,8 +1057,14 @@ chain with C<PerlOutputFilterHandler> and C<PerlSetOutputFilter>:
 
 Don't hesitate to contact me if you are interested in this little module.
 
+=head2 Debugging
+
+Sometimes it is useful to know the information encoded in a session
+identifier. This is why L<Apache::ClickPath::Decode> exists.
+
 =head1 SEE ALSO
 
+L<Apache::ClickPath::Decode(3)>
 L<http://perl.apache.org>,
 L<http://httpd.apache.org>
 
